/********************************************************************
*** instergen.c
*** written by Michael Thompson
***
*** This file is a utility program for the PDAsm tool (a Disassembler
*** for Parallax's Propeller chip.) When executed, this tool will
*** generate a C language source file defining an array which maps
*** Propeller opcodes to their corresponding propeller assembler
*** menemonics. The generated file is then compiled into the PDAsm
*** application to add support for new or missed mnemonics. 
********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>


/*** Array of 512 menmonic strings, up to 7 characters each      ***/
char gInstrName[512][8] = {0};

/********************************************************************
*** Converts an ASCII string of 1s and 0s into its binary equivilant.
***	Strings longer than 32 are truncated.
********************************************************************/
unsigned int BinStrToInt(const char * str)
{
	int c = 0;
	int ret = 0;
	unsigned int add = 1;

	for(c = strlen(str) - 1; c >= 0; --c)
	{
		if('1' == *(str + c))
			ret += add;

		add <<= 1;
	}

	return ret;
}

/********************************************************************
*** Reads a tab-delineated list of encoded-binary/mnemonic pairs into
*** an array of 512 entries using the encoded-binary number as a key.
********************************************************************/
void ImportInstrTable(FILE *file)
{
	char	bin[16] = { 0 };
	char	asm[8] = { 0 };

	//Read until "NOP" is encountered.
	while(1)
	{
		//Get the next encoded-binary/mnemonic pair.
		fscanf(file, "%s%s", bin, asm);

		// Break on "NOP" mnemonic
		if(0 == strcmp("NOP", asm))
			break;

		// Insert mnemonic at index given by value of encoded-binary.
		strncpy(gInstrName[BinStrToInt(bin)], asm, 7);
	}
}

/********************************************************************
*** Writes a C source file containing an array of 512 strings. Each
*** entry contains the mnemonic string where the index matches the
*** value of the encoded-binary.
********************************************************************/
void ExportInstrTable(FILE *file)
{
	int i = 0;

	/* Print out a header, some includes and open the array */
	fprintf(file, "/********************************************************************\n");
	fprintf(file, "*** instr.c\n");
	fprintf(file, "*** Generated by instergen.exe\n");
	fprintf(file, "***\n");
	fprintf(file, "*** This file defines an array which serves as a lookup table between\n");
	fprintf(file, "*** the identifying bits of an instruction WORD and its equivelant\n");
	fprintf(file, "*** assembly keyword.\n");
	fprintf(file, "********************************************************************/\n\n");
	fprintf(file, "#include \"instr.h\"\n\n\n");
	fprintf(file, "char gInstrName[512][8] = {\n");
	
	/* Print out the 512 possible mnemonics. */
	for(i = 0; i < 511; ++i)
	{
		if(0 == (i % 16))
		{
			/* Every 16 lines, print a comment with the current index of the array. */
			fprintf(file, "/* %3.3d */\t", i);
		}
		else
		{
			/* Otherwise just tab out to the appropriate depth. */
			fprintf(file, "\t\t\t");
		}

		/* Print out the mnemonic. */
		fprintf(file, "\"%s\",\n", gInstrName[i]);
	}

	/* Close the array and print the footer */
	fprintf(file, "\t\t\t\"%s\"\n};\n\n", gInstrName[i]);
	fprintf(file, "/* EOF */");
}


/********************************************************************
*** Opens pre-defined files for input and output, imports data from
*** the input file 'instr.txt', exports data to the file 'instr.c'
*** and finally closes the files.
********************************************************************/
int main(int argc, char **argv)
{
	FILE *	out = 0;
	FILE *	in = 0;

	// open and make sure we have a valid FILE * for input
	if(!(in = fopen("instr.txt", "r")))
	{
		printf("Error opening file: instr.txt");
		return 1;
	}

	// open and make sure we have a valid FILE * for output
	if(!(out = fopen("instr.c", "w")))
	{
		printf("Error opening file: instr.c");
		return 1;
	}

	ImportInstrTable(in);
	ExportInstrTable(out);

	fclose(out);
	fclose(in);
}

/* EOF */